from datetime import date, datetime


def current_datetime() -> datetime:
    """Returns the current datetime"""
    return datetime.utcnow()


def current_date() -> date:
    """Returns the current datetime"""
    return current_datetime().date()


def current_timestamp() -> int:
    """
    Returns the current timestamp from epoch (rounded to the nearest second)
    """
    return int(datetime.timestamp(current_datetime()))



import glob
import os
from typing import List, Optional, Set, Tuple


def explode(path: str) -> Tuple[str, str, str]:
    """
    Split a file path into 3 parts:
    - Head (directory)
    - File Name
    - Extension (without dot '.')
    """
    head, tail = os.path.split(path)
    name, extension = os.path.splitext(tail)
    extension = extension.lstrip(".")
    return head, name, extension


def search_files(
    directory: str,
    *,
    filter_endswith: Optional[str] = None,
    filter_extensions: Optional[Set[str]] = None,
    does_not_contain: Optional[Set[str]] = None,
) -> List[str]:
    """Retrieve files in a directory, matching given criteria"""

    def _does_not_contain(path: str) -> bool:
        if does_not_contain is None:
            return True
        _, name, _ = explode(path)
        return not any([(item in name) for item in does_not_contain])

    def _endswith(path: str) -> bool:
        if filter_endswith is None:
            return True
        _, name, _ = explode(path)
        pattern = filter_endswith.lower()
        return name.endswith(pattern)

    def _extension(path: str) -> bool:
        if filter_extensions is None:
            return True
        _, _, ext = explode(path)
        if not ext:
            return False
        return ext.lower().strip(".") in filter_extensions

    all_files = glob.glob(directory + "/*")

    filtered = filter(_does_not_contain, all_files)
    filtered = filter(_endswith, filtered)
    filtered = filter(_extension, filtered)

    return list(filtered)



from typing import Any, Callable, Dict, List, Mapping, Sequence, Union

EntitiesType = List[Dict]

# https://stackoverflow.com/questions/51291722/define-a-jsonable-type-using-mypy-pep-526
JsonType = Union[Sequence, Mapping, str, int, float, bool]

Callback = Callable[[Any], Any]
Getter = Union[str, Callback]



"""
Request timeout in seconds for Looker API
"""
DEFAULT_LOOKER_TIMEOUT_SECOND = 120
CASTOR_LOOKER_TIMEOUT_SECOND = "CASTOR_LOOKER_TIMEOUT_SECOND"

"""
Number of items per page when requesting Looker API
"""
DEFAULT_LOOKER_PAGE_SIZE = 500
CASTOR_LOOKER_PAGE_SIZE = "CASTOR_LOOKER_PAGE_SIZE"



from typing import Dict, Iterator, Optional

from ....utils import EntitiesType
from ..entities import EXPORTED_FIELDS, QUERY_FIELDS, TableauEntity

QUERY_TEMPLATE = """
{{
  {object_type}Connection(first: {page_size}, after: AFTER_TOKEN_SIGNAL) {{
    nodes {{ {query_fields}
    }}
    pageInfo {{
      hasNextPage
      endCursor
    }}
    totalCount
  }}
}}
"""

RESOURCE_TEMPLATE = "{resource}Connection"


def get_paginated_objects(
    server, entity: TableauEntity, page_size: int
) -> EntitiesType:

    fields = QUERY_FIELDS[entity]["fields"].value
    object_type = QUERY_FIELDS[entity]["object_type"].value
    query = QUERY_TEMPLATE.format(
        object_type=object_type,
        page_size=page_size,
        query_fields=fields,
    )
    resource = RESOURCE_TEMPLATE.format(resource=object_type)

    return [
        result
        for results in query_scroll(server, query, resource)
        for result in results
    ]


def query_scroll(server, query: str, resource: str) -> Iterator[EntitiesType]:
    """build a tableau query iterator handling pagination and cursor"""

    def _call(cursor: Optional[str]) -> dict:
        # If cursor is defined it must be quoted else use null token
        token = "null" if cursor is None else f'"{cursor}"'
        query_ = query.replace("AFTER_TOKEN_SIGNAL", token)

        return server.metadata.query(query_)["data"][resource]

    cursor = None
    while True:
        payload = _call(cursor)
        yield payload["nodes"]

        page_info = payload["pageInfo"]
        if page_info["hasNextPage"]:
            cursor = page_info["endCursor"]
        else:
            break


def extract_entity(entity: Dict, entity_type: TableauEntity) -> Dict:
    """Agnostic function extracting dedicated attributes with define entity"""
    return {key: getattr(entity, key) for key in EXPORTED_FIELDS[entity_type]}



import logging
from typing import Iterable, Tuple, Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .client import ApiClient, DbClient
from .entities import MetabaseEntity

logger = logging.getLogger(__name__)

ClientMetabase = Union[DbClient, ApiClient]


def iterate_all_data(
    client: ClientMetabase,
) -> Iterable[Tuple[MetabaseEntity, list]]:
    """Iterate over the extracted Data From metabase"""

    yield MetabaseEntity.USER, deep_serialize(client.fetch(MetabaseEntity.USER))
    yield MetabaseEntity.COLLECTION, deep_serialize(
        client.fetch(MetabaseEntity.COLLECTION)
    )
    yield MetabaseEntity.DATABASE, deep_serialize(
        client.fetch(MetabaseEntity.DATABASE)
    )
    yield MetabaseEntity.TABLE, deep_serialize(
        client.fetch(MetabaseEntity.TABLE)
    )
    yield MetabaseEntity.CARD, deep_serialize(client.fetch(MetabaseEntity.CARD))
    yield MetabaseEntity.DASHBOARD, deep_serialize(
        client.fetch(MetabaseEntity.DASHBOARD)
    )
    yield MetabaseEntity.DASHBOARD_CARDS, deep_serialize(
        client.fetch(MetabaseEntity.DASHBOARD_CARDS)
    )


def extract_all(client: ClientMetabase, **kwargs: str) -> None:
    """
    Extract Data From metabase
    Store the output files locally under the given output_directory
    """
    output_directory = kwargs.get("output_directory") or from_env(OUTPUT_DIR)
    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.name.lower(), output_directory, ts)
        write_json(filename, data)

    write_summary(
        output_directory,
        ts,
        base_url=client.base_url(),
        client_name=client.name(),
    )



from typing import Any, DefaultDict, Dict, List, Sequence

from .object import getproperty
from .type import Getter


def group_by(identifier: Getter, elements: Sequence) -> Dict[Any, List]:
    """Groups the elements by the given key"""
    groups: Dict[Any, List] = DefaultDict(list)
    for element in elements:
        key = getproperty(element, identifier)
        groups[key].append(element)

    return groups



from typing import List, Optional

from ..abstract import (
    QUERIES_DIR,
    AbstractQueryBuilder,
    ExtractionQuery,
    TimeFilter,
    WarehouseAsset,
)

# Those queries must be formatted with {region}
from .typing import IterTwoString

REGION_REQUIRED = (
    WarehouseAsset.COLUMN,
    WarehouseAsset.DATABASE,
    WarehouseAsset.QUERY,
    WarehouseAsset.SCHEMA,
    WarehouseAsset.TABLE,
    WarehouseAsset.USER,
)

# Those queries must be formatted with {dataset}
DATASET_REQUIRED = (WarehouseAsset.VIEW_DDL,)

# Those queries must be de-duplicated
# The usage of DISTINCT in the query is not enough
# because we stitch several queries results
BIGQUERY_DUPLICATES = (
    WarehouseAsset.DATABASE,
    WarehouseAsset.USER,
)

SHARDED_ASSETS = (WarehouseAsset.TABLE, WarehouseAsset.COLUMN)
SHARDED_FILE_PATH = "cte/sharded.sql"


def _database_formatted(datasets: IterTwoString) -> str:
    databases = {db for _, db in datasets}
    if not databases:
        # when no datasets are provided condition should pass
        return "(NULL)"
    formatted = ", ".join([f"'{db}'" for db in databases])
    return f"({formatted})"


class BigQueryQueryBuilder(AbstractQueryBuilder):
    """
    Builds queries to extract assets from BigQuery.
    Generate multiple queries to support multi-regions
    """

    def __init__(
        self,
        regions: IterTwoString,
        datasets: IterTwoString,
        time_filter: Optional[TimeFilter] = None,
        sync_tags: Optional[bool] = False,
    ):
        super().__init__(
            time_filter=time_filter, duplicated=BIGQUERY_DUPLICATES
        )
        self._regions = regions
        self._datasets = datasets
        self._sync_tags = sync_tags

    @staticmethod
    def _format(query: ExtractionQuery, values: dict) -> ExtractionQuery:
        return ExtractionQuery(
            statement=query.statement.format(**values),
            params=query.params,
        )

    def file_name(self, asset: WarehouseAsset) -> str:
        """
        Returns the SQL filename extracting the given asset.
        Overrides the default behaviour - handle table tags for BigQuery
        """
        if asset == WarehouseAsset.TABLE and self._sync_tags:
            # Reading `INFORMATION_SCHEMA.SCHEMATA_OPTIONS` requires specific permissions.
            # Synchronization of tags is only activated when credentials are sufficient.
            return f"{asset.value}_with_tags.sql"

        return f"{asset.value}.sql"

    def load_statement(self, asset: WarehouseAsset) -> str:
        """load sql statement from file"""
        statement = super().load_statement(asset)

        if asset not in SHARDED_ASSETS:
            return statement

        sharded_statement = self._load_from_file(SHARDED_FILE_PATH)
        return statement.format(sharded_statement=sharded_statement)

    def build(self, asset: WarehouseAsset) -> List[ExtractionQuery]:
        """
        It would be easier to stitch data directly in the query statement (UNION ALL).
        Unfortunately, querying INFORMATION_SCHEMA on multiple regions
          at the same time gives partial result (seems like a BigQuery bug)
        This weird behaviour forces us to
          - run one query per tuple(project, region)
          -  stitch data afterwards.
        """
        query = super().build_default(asset)

        if asset in REGION_REQUIRED:
            return [
                self._format(query, {"project": project, "region": region})
                for project, region in self._regions
            ]

        if asset in DATASET_REQUIRED:
            return [
                self._format(query, {"project": project, "dataset": dataset})
                for project, dataset in self._datasets
            ]

        return [query]



import logging
from argparse import ArgumentParser

from castor_extractor.visualization import tableau  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    user_group = parser.add_mutually_exclusive_group(required=False)
    user_group.add_argument("-u", "--user", help="Tableau user")
    user_group.add_argument("-n", "--token-name", help="Tableau token name")

    password_group = parser.add_mutually_exclusive_group(required=False)
    password_group.add_argument("-p", "--password", help="Tableau password")
    password_group.add_argument("-t", "--token", help="Tableau token")

    parser.add_argument("-b", "--server-url", help="Tableau server url")
    parser.add_argument("-i", "--site-id", help="Tableau site ID")
    parser.add_argument(
        "-s", "--safe-mode", help="Tableau safe mode", action="store_true"
    )
    parser.add_argument("-o", "--output", help="Directory to write to")

    args = parser.parse_args()

    client = tableau.ApiClient(
        user=args.user,
        password=args.password,
        token_name=args.token_name,
        token=args.token,
        server_url=args.server_url,
        site_id=args.site_id,
        safe_mode=args.safe_mode,
    )

    client.login()

    tableau.extract_all(
        client,
        output_directory=args.output,
    )
